"""
Module to get schemas from `Chimera's` registry.
"""
import requests
from chimera_client.environ import Environment


class Schema:
    """
    Class to get schemas from `Chimera`'s registry.
    """

    def getSchema(self, channel: str) -> str:
        """
        Gets the schema from a given channel.
        """
        env = Environment()
        r = requests.get(env.chimeraRegistryUrl + '/schema/' +
                         env.chimeraNamespace + '/' + channel)
        if r.status_code != 200:
            raise Exception('[GET SCHEMA] Unable to retrieve schema.')

        # Convert body from byte to string
        return r.content.decode('utf-8')



"""
Module that gets all envorinment variables from the context as to allow chimera communication.
"""
import os

class Environment:
    """
    A class that gets all necessary variables from the context for chimera commuication.
    """

    def __init__(self):
        try:
            self.chimeraNodeID: str = os.environ['CHIMERA_NODE_ID']
            self.chimeraNamespace: str = os.environ['CHIMERA_NAMESPACE']
            self.chimeraRegistryUrl: str = os.environ['CHIMERA_REGISTRY_URL']
            self.outputChannels: str = os.environ['CHIMERA_OUTPUT_CHANNELS']
            self.inputChannels: str = os.environ['CHIMERA_INPUT_CHANNELS']

            self.bootstrapServers: str = os.environ['KAFKA_BOOTSTRAP_SERVERS']
        except KeyError:
            raise EnvironmentError('Check the environment vars.')



"""
Module containing the reader for chimera.
"""

import os
import io
import avro.io
import avro.schema
from typing import List
from confluent_kafka import Message, Consumer
from chimera_client.environ import Environment
from chimera_client.schema import Schema
from typing import Any, Dict


class ChimeraConsumer:

    def __init__(self, channels):
        environ = Environment()
        self._consumer = Consumer({
            'bootstrap.servers': environ.bootstrapServers,
            'group.id': environ.chimeraNodeID,
            'auto.offset.reset': "earliest",
            'enable.auto.commit': False,
        })

        self._consumer.subscribe(channels)
        self._lastMessage: Message = None


class Consumers(dict):
    def __getitem__(self, key):

        # Get the specific channel consumer of Pipeline
        if key:
            return super().__getitem__(key)
        return super().__getitem__('all')
    def set_last_message(self, channel: str, message: Message):
        super().__getitem__(channel)._lastMessage = message


class Reader:
    """
    Reader is a class that reads messages from `Chimera`'s channels, and decodes them.
    """

    def __init__(self, readAll: bool = True):
        """
        Initializes the reader with all information needed.

        Parameter
            readAll: bool 

            True: Read messages comming from all channels.
            Call readMessage() with no parameters.

            False: Read messages from a scpecific channel.
            Call readMessage(channel) where channel is the desired channel.

        """
        environ = Environment()
        channels: List[str] = list(map(
            lambda x: environ.chimeraNamespace + "_" + x, environ.inputChannels.split(';')[:-1]))

        self.__consumers = Consumers()
        if readAll:
            self.__consumers['all'] = ChimeraConsumer(channels)
        else:
            # Create a consumer for each channel
            for channel in channels:
                self.__consumers[channel] = ChimeraConsumer(channel)

    def readMessage(self, channel: str = None) -> Any:
        """
        Reads a message from chimera.

        Returns the message read from channel.
        """

        

        while True:
            messageEncoded: Message = self.__consumers[channel]._consumer.poll(100)
            if messageEncoded is None:
                continue

            if messageEncoded.error():
                raise Exception('[READ MESSAGE]: ', messageEncoded.error())

            self.__consumers[channel]._lastMessage = messageEncoded
            try:
                chan = "_".join(messageEncoded.topic().split('_')[1:])
                # Decoding  message
                bytes_reader = io.BytesIO(messageEncoded.value())
                decoder = avro.io.BinaryDecoder(bytes_reader)
                schema = avro.schema.Parse(
                    Schema().getSchema(chan))
                reader = avro.io.DatumReader(schema)
                message = reader.read(decoder)

            except Exception as e:
                print("Couldn't desserialize message: ", message.value())
                print(e)
                continue

            return message

    def commit(self, channel: str = None):
        """
        Commits the last message received.
        """

        self.__consumers[channel]._consumer.commit(self.__consumers[channel]._lastMessage)

    def close(self, channel: str = None):
        """
        Closes connection with the chimera cluster.
        """

        self.__consumers[channel]._consumer.close()



from setuptools import setup
from os import path


# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='chimera_client',
    packages=['chimera_client'],
    description='Client for communication using chimera channels.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.0.2',
    url="https://gitlab.inspr.com/chimera/client",
    install_requires=['requests', 'avro-python3', 'confluent_kafka']
)



"""
Module containing the `Chimera` writer class.
"""

import io
import os
import avro.io
import avro.schema
from typing import List
from confluent_kafka import Producer
from chimera_client.environ import Environment
from chimera_client.schema import Schema

class Writer:
    """
    Writer is a class that writes into chimera. 
    """

    def __init__(self):
        """
        Initializes the writer with the environment variables.
        """
        self.__environ = Environment()
        self._producer: Producer = Producer({'bootstrap.servers': self.__environ.bootstrapServers})


    def writeMessage(self, message: object, channel: str):
        """
        Writes a message in the given `Chimera` topic.

        The message should be in the format specified in the channel's avro schema, set previously.

        - message: the message that will be encoded.
        - channel: the channel on which the message will be sent

        """
        allowedTopics: List[str] = self.__environ.outputChannels.split(';')
        print(allowedTopics)
        if channel not in allowedTopics:
            raise Exception('Not allowed to write on channel ' + channel + '.')

        # Encoding message
        json = Schema().getSchema(channel)
        print(json)
        schema = avro.schema.Parse(json)
        avrowriter = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        avrowriter.write(message, encoder)

        value = bytes_writer.getvalue()

        self._producer.produce(
            topic=self.__environ.chimeraNamespace + "_" + channel, value=value)
        self._producer.flush()
